==> simply.core-1.1.93.jar: simply/gcp/auth.clj <==
(ns simply.gcp.auth
  (:require [simply.http.client :as http]
            [simply.errors :as e]
            [integrant.core :as ig]
            [clojure.string :as string])
  (:import com.google.api.client.googleapis.auth.oauth2.GoogleCredential
           java.io.FileInputStream
           java.util.Collections))


;;;; CREDENTIALS

(def *scoped-credentials (atom {}))


(defn reset-credentials! []
  (reset! *scoped-credentials {}))


(defn- scoped-token
  "returns an existing oauth token, or fetches one if expired"
  [credential scope]
  (when-not (contains? @*scoped-credentials scope)
    (swap! *scoped-credentials assoc scope credential))
  (let [cred (get @*scoped-credentials scope)]
    (cond-> cred
      (nil? cred) (e/throw-app-error "Google Credential Cannot be Nil" {:scope scope})
      (nil? (.getAccessToken cred)) .refreshToken
      (> 60 (.getExpiresInSeconds cred)) .refreshToken)
    (swap! *scoped-credentials assoc scope cred)
    (.getAccessToken cred)))


(defn- local-credential [scope keyfile-path]
  (-> (FileInputStream. keyfile-path)
      GoogleCredential/fromStream
      (.createScoped (Collections/singleton scope))))


;;;; CLIENTS

(def ^:private datastore-api-endpoint  "https://datastore.googleapis.com/v1/projects/")

(defn- re-route-to-emulator-if-datastore
  "We do this here cause sometimes a hack is much much less work"
  [req {:keys [datastore-port datastore-host]
        :or {datastore-port 8080 datastore-host "localhost"}}]
  (if (string/starts-with? (:url req) datastore-api-endpoint)
    (let [splits (string/split (:url req) #":" 3)
          replace-project? (= 3 (count splits))
          endpoint (str "http://" datastore-host ":" datastore-port "/v1/projects/")]
      (if replace-project?
        (assoc req :url (str endpoint "dogmatix:" (last splits)))
        (update req :url string/replace-first "https://datastore.googleapis.com/v1/projects/" endpoint)))
    req))


(defn emulator-client
  "For use with datastore and pubsub emulators that run locally."
  [& {:keys [datastore-port datastore-host] :as opts}]
  (fn [req]
    (-> req
        (re-route-to-emulator-if-datastore opts)
        http/request)))


(defn- credentials-client
  "Used to access GCP api's while authenticating using a json keyfile."
  [cred {:keys [scope] :as request}]
  (http/request
   (assoc request
          :oauth-token (scoped-token cred scope))))


(defn keyfile-client [keyfile-path]
  (fn [{:keys [scope] :as request}]
    (credentials-client (local-credential scope keyfile-path) request)))


(defn default-credentials-client
  "Used by containers running on google cloud that have access to default credentials."
  []
  (fn [request]
    (credentials-client (GoogleCredential/getApplicationDefault) request)))


;;;; INTEGRANT

(defmethod ig/init-key :simply.gcp.auth/emulator-client [_ opts] (apply emulator-client opts))


(defmethod ig/init-key :simply.gcp.auth/keyfile-client
  [_ {:keys [keyfile-path]}]
  (keyfile-client keyfile-path))


(defmethod ig/init-key :simply.gcp.auth/default-credentials-client [_ _]
  (default-credentials-client))

==> simply.core-1.1.117.jar: simply/gcp/auth.clj <==
(ns simply.gcp.auth
  (:require [simply.http.client :as http]
            [simply.errors :as e]
            [integrant.core :as ig]
            [clojure.string :as string]
            [cheshire.core :as json])
  (:import com.google.api.client.googleapis.auth.oauth2.GoogleCredential
           java.io.FileInputStream
           java.util.Collections
           [com.google.auth.oauth2 GoogleCredentials UserCredentials]))


;;;; CREDENTIALS

(def *scoped-credentials (atom {}))


(defn reset-credentials! []
  (reset! *scoped-credentials {}))


(defmulti refresh type)
(defmulti get-access-token type)


(defmethod refresh UserCredentials [cred]
  (if (nil? (.getAccessToken cred))
    (.refresh cred)
    (.refreshIfExpired cred)))

(defmethod get-access-token UserCredentials [cred]
  (.getTokenValue (.getAccessToken cred)))


(defmethod refresh :default [cred]
  (cond-> cred
    (nil? (.getAccessToken cred))      .refreshToken
    (> 60 (.getExpiresInSeconds cred)) .refreshToken))

(defmethod get-access-token :default [cred]
  (.getAccessToken cred))


(defn- scoped-token
  "returns an existing oauth token, or fetches one if expired"
  [credential scope]
  (when-not (contains? @*scoped-credentials scope)
    (swap! *scoped-credentials assoc scope credential))
  (let [cred (get @*scoped-credentials scope)]
    (if (nil? cred)
      (e/throw-app-error "Google Credential Cannot be Nil" {:scope scope})
      (refresh cred))
    (swap! *scoped-credentials assoc scope cred)
    (get-access-token cred)))


(defn- local-credential [scope keyfile-path]
  (-> (FileInputStream. keyfile-path)
      GoogleCredential/fromStream
      (.createScoped (Collections/singleton scope))))


;;;; CLIENTS

(def ^:private datastore-api-endpoint  "https://datastore.googleapis.com/v1/projects/")

(defn- re-route-to-emulator-if-datastore
  "We do this here cause sometimes a hack is much much less work"
  [req {:keys [datastore-port datastore-host]
        :or {datastore-port 8080 datastore-host "localhost"}}]
  (if (string/starts-with? (:url req) datastore-api-endpoint)
    (let [splits (string/split (:url req) #":" 3)
          replace-project? (= 3 (count splits))
          endpoint (str "http://" datastore-host ":" datastore-port "/v1/projects/")]
      (if replace-project?
        (assoc req :url (str endpoint "dogmatix:" (last splits)))
        (update req :url string/replace-first "https://datastore.googleapis.com/v1/projects/" endpoint)))
    req))


(defn emulator-client
  "For use with datastore and pubsub emulators that run locally."
  [& {:keys [datastore-port datastore-host] :as opts}]
  (fn [req]
    (-> req
        (re-route-to-emulator-if-datastore opts)
        http/request)))


(defn- credentials-client
  "Used to access GCP api's while authenticating using a json keyfile."
  [cred {:keys [scope] :as request}]
  (http/request
   (assoc request
          :oauth-token (scoped-token cred scope))))


(defn ^{:deprecated "1.1.116"} keyfile-client [keyfile-path]
  (fn [{:keys [scope] :as request}]
    (credentials-client (local-credential scope keyfile-path) request)))


(defn default-credentials-client
  "Used by containers running on google cloud that have access to default credentials."
  []
  (fn [request]
    (credentials-client (GoogleCredential/getApplicationDefault) request)))


(defn developer-credential []
  (GoogleCredentials/getApplicationDefault))


(defn developer-client
  []
  (fn [request]
    (credentials-client (developer-credential) request)))


(defn developer-email []
  (let [client  (developer-client)
        response @(client
                    {:method :get
                     :url    "https://www.googleapis.com/oauth2/v1/userinfo"})
        status   (:status response)
        body     (-> response :body (json/parse-string keyword))]
    (if (= 200 status)
      (:email body)
      (throw (ex-info "Have you run `gcloud auth application-default login`"
                      {:status status :body body})))))


;;;; INTEGRANT

(defmethod ig/init-key :simply.gcp.auth/emulator-client [_ opts] (apply emulator-client opts))


(defmethod ig/init-key :simply.gcp.auth/keyfile-client
  [_ {:keys [keyfile-path]}]
  (keyfile-client keyfile-path))


(defmethod ig/init-key :simply.gcp.auth/default-credentials-client [_ _]
  (default-credentials-client))


(defmethod ig/init-key :simply.gcp.auth/developer-credentials-client
  [_ _]
  (developer-client))
